
import 'dart:async';

import 'package:flutter/foundation.dart';
import 'package:flutter/gestures.dart';
import 'package:flutter/services.dart';
import 'package:flutter/widgets.dart';

typedef void WebViewCreatedCallback(BridgeWebViewController controller);

enum JavascriptMode {
    disabled,
    unrestricted,
}

class NavigationRequest {
    NavigationRequest._({this.url, this.isForMainFrame});

    /// The URL that will be loaded if the navigation is executed.
    final String url;
    /// Whether the navigation request is to be loaded as the main frame.
    final bool isForMainFrame;

    @override
    String toString() {
        return '$runtimeType(url: $url, isForMainFrame: $isForMainFrame)';
    }
}

enum NavigationDecision {
    /// Prevent the navigation from taking place.
    prevent,
    /// Allow the navigation to take place.
    navigate,
}

typedef NavigationDecision NavigationDelegate(NavigationRequest navigation);

typedef void PageFinishedCallback(String url);

class BridgeWebView extends StatefulWidget {
    const BridgeWebView({
        Key key,
        this.onWebViewCreated,
        this.initialUrl,
        this.javascriptMode = JavascriptMode.disabled,
        this.navigationDelegate,
        this.gestureRecognizers,
        this.onPageFinished,
    })
        : assert(javascriptMode != null),
            super(key: key);

    final WebViewCreatedCallback onWebViewCreated;
    final Set<Factory<OneSequenceGestureRecognizer>> gestureRecognizers;
    final String initialUrl;
    final JavascriptMode javascriptMode;
    final NavigationDelegate navigationDelegate;
    final PageFinishedCallback onPageFinished;

    @override
    State<StatefulWidget> createState() => _BridgeWebViewState();
}

class _BridgeWebViewState extends State<BridgeWebView> {
    final Completer<BridgeWebViewController> _controller =
    Completer<BridgeWebViewController>();

    _WebSettings _settings;

    @override
    Widget build(BuildContext context) {
        if (defaultTargetPlatform == TargetPlatform.android) {
            return GestureDetector(
                onLongPress: () {},
                excludeFromSemantics: true,
                child: AndroidView(
                    viewType: 'com.microduino.flutter/bridgewebview',
                    onPlatformViewCreated: _onPlatformViewCreated,
                    gestureRecognizers: widget.gestureRecognizers,
                    // WebView content is not affected by the Android view's layout direction,
                    // we explicitly set it here so that the widget doesn't require an ambient
                    // directionality.
                    layoutDirection: TextDirection.rtl,
                    creationParams: _CreationParams.fromWidget(widget).toMap(),
                    creationParamsCodec: const StandardMessageCodec(),
                ),
            );
        } else if (defaultTargetPlatform == TargetPlatform.iOS) {
            return UiKitView(
                viewType: 'com.microduino.flutter/bridgewebview',
                onPlatformViewCreated: _onPlatformViewCreated,
                gestureRecognizers: widget.gestureRecognizers,
                creationParams: _CreationParams.fromWidget(widget).toMap(),
                creationParamsCodec: const StandardMessageCodec(),
            );
        }
        return Text(
            '$defaultTargetPlatform is not yet supported by the webview_flutter plugin');
    }

    @override
    void initState() {
        super.initState();
    }

    @override
    void didUpdateWidget(BridgeWebView oldWidget) {
        super.didUpdateWidget(oldWidget);
        _updateConfiguration(_WebSettings.fromWidget(widget));
    }

    Future<void> _updateConfiguration(_WebSettings settings) async {
        _settings = settings;
        final BridgeWebViewController controller = await _controller.future;
        controller._updateSettings(settings);
        controller._navigationDelegate = widget.navigationDelegate;
    }

    void _onPlatformViewCreated(int id) {
        final BridgeWebViewController controller = BridgeWebViewController._(
            id,
            _WebSettings.fromWidget(widget),
            widget.navigationDelegate,
            widget.onPageFinished,
        );
        _controller.complete(controller);
        if (widget.onWebViewCreated != null) {
            widget.onWebViewCreated(controller);
        }
    }

}

class _CreationParams {
    _CreationParams(
        {this.initialUrl, this.settings});

    static _CreationParams fromWidget(BridgeWebView widget) {
        return _CreationParams(
            initialUrl: widget.initialUrl,
            settings: _WebSettings.fromWidget(widget),
        );
    }

    final String initialUrl;
    final _WebSettings settings;

    Map<String, dynamic> toMap() {
        return <String, dynamic>{
            'initialUrl': initialUrl,
            'settings': settings.toMap()
        };
    }
}

class _WebSettings {
    _WebSettings({
        this.javascriptMode,
        this.hasNavigationDelegate,
    });

    static _WebSettings fromWidget(BridgeWebView widget) {
        return _WebSettings(
            javascriptMode: widget.javascriptMode,
            hasNavigationDelegate: widget.navigationDelegate != null,
        );
    }

    final JavascriptMode javascriptMode;
    final bool hasNavigationDelegate;

    Map<String, dynamic> toMap() {
        return <String, dynamic>{
            'jsMode': javascriptMode.index,
            'hasNavigationDelegate': hasNavigationDelegate,
        };
    }

    Map<String, dynamic> updatesMap(_WebSettings newSettings) {
        final Map<String, dynamic> updates = <String, dynamic>{};
        if (javascriptMode != newSettings.javascriptMode) {
            updates['jsMode'] = newSettings.javascriptMode.index;
        }
        if (hasNavigationDelegate != newSettings.hasNavigationDelegate) {
            updates['hasNavigationDelegate'] =
                newSettings.hasNavigationDelegate;
        }
        return updates;
    }
}

class BridgeWebViewController {
    BridgeWebViewController._(int id,
        this._settings,
        this._navigationDelegate,
        this._onPageFinished,)
        : _channel = MethodChannel('com.microduino.flutter/bridgewebview_$id') {
        _channel.setMethodCallHandler(_onMethodCall);
    }

    final MethodChannel _channel;

    NavigationDelegate _navigationDelegate;

    _WebSettings _settings;

    final PageFinishedCallback _onPageFinished;

    Future<bool> _onMethodCall(MethodCall call) async {
        switch (call.method) {
            case 'navigationRequest':
                final NavigationRequest request = NavigationRequest._(
                    url: call.arguments['url'],
                    isForMainFrame: call.arguments['isForMainFrame'],
                );
                final bool allowNavigation = _navigationDelegate == null ||
                    _navigationDelegate(request) == NavigationDecision.navigate;
                return allowNavigation;
            case 'onPageFinished':
                if (_onPageFinished != null) {
                    _onPageFinished(call.arguments['url']);
                }

                return null;
        }
        throw MissingPluginException(
            '${call.method} was invoked but has no handler');
    }

    Future<void> loadUrl(String url) async {
        assert(url != null);
        _validateUrlString(url);
        return _channel.invokeMethod('loadUrl', url);
    }

    Future<void> callHandler(String method) async {
        assert(method != null);
        _validateUrlString(method);
        return _channel.invokeMethod('callHandler', method);
    }

    Future<void> registerHandler(String method) async {
        assert(method != null);
        _validateUrlString(method);
        return _channel.invokeMethod('registerHandler', method);
    }

    Future<String> currentUrl() async {
        final String url = await _channel.invokeMethod('currentUrl');
        return url;
    }

    Future<bool> canGoBack() async {
        final bool canGoBack = await _channel.invokeMethod("canGoBack");
        return canGoBack;
    }

    Future<bool> canGoForward() async {
        final bool canGoForward = await _channel.invokeMethod("canGoForward");
        return canGoForward;
    }

    Future<void> goBack() async {
        return _channel.invokeMethod("goBack");
    }

    Future<void> goForward() async {
        return _channel.invokeMethod("goForward");
    }

    Future<void> reload() async {
        return _channel.invokeMethod("reload");
    }

    Future<void> clearCache() async {
        await _channel.invokeMethod("clearCache");
        return reload();
    }

    Future<void> _updateSettings(_WebSettings setting) async {
        final Map<String, dynamic> updateMap = _settings.updatesMap(setting);
        if (updateMap == null || updateMap.isEmpty) {
            return null;
        }
        _settings = setting;
        return _channel.invokeMethod('updateSettings', updateMap);
    }

}

void _validateUrlString(String url) {
    try {
        final Uri uri = Uri.parse(url);
        if (uri.scheme.isEmpty) {
            throw ArgumentError('Missing scheme in URL string: "$url"');
        }
    } on FormatException catch (e) {
        throw ArgumentError(e);
    }
}
